function [ akmat, lkmat, fnest, err_form, titer ] = form2af( ftarg, a0, l0, dl, norm_lips, carea, cleng, cpot, thresh, Nbreak, param )
% Dynamic acoustic-to-articulatory inversion of oral vowels using the sensitivity
% matrix technique
%
% Input arguments:
%
% ftarg : formant frequencies. Should be a matrix where each column contains the formant frequencies of a temporal frame
% a0 : initial area function (seed function)
% l0 : inital length function
% dl : choice of the length variation (default = 0)
% norm_lips : constraints on the lips opening (default = 0, no constraint)
% carea : weight of area constraint (default = 0)
% cleng : weight of length constraint (default = 0)
% cpot : weight of the potential energy constraint (default = 0)
% thresh : threshold (default = 0.01)
% Nbreak : maximum number of iteration (default = 1000)
% param : parameters for the chain matrix model (default = Sondhi and Schroeter parameters)
% disp : choice of display (default = 0, no display)
%
% Output arguments
%
% akmat : matrix containing the area function of each temporal frame (from the glottis to the lips)
% lkmat : matrix containing the length functions of each temporal frame (from the glottis to the lips) 
% fnest : matrix containing the acoustic vector of each temporal frame
% err_form : vector contain,ing the value of the cost function at each iteration
% titer : vector containing the elapsed time at each iteration
%
% Benjamin Elie January 2013
%
% !! THE AREA FUNCTION SHOULD BE FROM THE LIPS TO THE GLOTTIS !!

if nargin < 4; dl = 0; end;
if nargin < 5; norm_lips = 0; end;
if nargin < 6; carea = 0; end;
if nargin < 7; cleng = 0; end;
if nargin < 8; cpot = 0; end;
if nargin < 9; thresh = 0.01; end;
if nargin < 10; Nbreak = 1000; end;
if nargin < 11; run('Library/constantterms.m'); end;

tic

Nft = size(ftarg,1); % size of the acoustic vector
Ntrame = size(ftarg,2); % number of frames

% Vector must be column vectors
a0vec = a0(:);
l0vec = l0(:);
ftargvec = ftarg(:);

% Initalization
freq = param.freq;
count_loop = 1;
ak = a0vec; akp1 = ak;
lk = l0vec; lkp1 = lk;
akp1mat = a0; lkp1mat = l0;
fnest = [];
Sn_xmat = zeros(length(a0vec),length(ftargvec));
Sn_xlmat = Sn_xmat;
Nt = size(a0,1); % size of the input vector (size of the area function)

if Ntrame == 1; carea = 0; cleng = 0; end; % if only one frame, no kinetic energy constraints (static configuration)

for k = 1:Ntrame % Seed resonance frequencies

    fk = af2form( a0(:,k), l0(:,k), freq, param ); % Compute resonance frequencies at current iteration 
    fn = fk(1:min(length(fk),Nft)); fnest = [fnest; fn(:)]; % Keep the Nft first resonance frequencies    
    Sn_x  = Sensitivity_function_fn_v2( l0(:,k), a0(:,k), fn, param ); % Compute the sensitivity matrix for the area function
      
    if dl
 
        Sn_x_l  = Sensitivity_function_length_fn_v2( l0(:,k), a0(:,k), fn, param); % Compute the sensitivity matrix for the length function if length estimation is enabled
        Sn_xlmat((k-1)*Nt+1:k*Nt,(k-1)*Nft+1:k*Nft) = Sn_x_l; 
    
    end
    
    Sn_xmat((k-1)*Nt+1:k*Nt,(k-1)*Nft+1:k*Nft) = Sn_x;    

end

err_form(count_loop) = norm((fnest-ftargvec)./fnest,1)*100; % distance from the target acoustic vector (cost function)
titer(count_loop) = toc;

while (err_form(end) >= thresh & count_loop <= Nbreak) % as long as the distance is higher than thresh
        
    zn = (ftargvec-fnest)./fnest; zn = zn(:); % Df vector
    
    % Acceleration term. psii > 1 => enhance speed, but risk of unstability, psii < 1 to make sure we are in the linearity domain...
    psii = 0.05;
    alph = real(transpose(zn)*transpose(Sn_xmat)*Sn_xmat*zn/(transpose(transpose(Sn_xmat)*Sn_xmat*zn)*(transpose(Sn_xmat)*Sn_xmat*zn)))*psii; 
    alphl = real(transpose(zn)*transpose(Sn_xlmat)*Sn_xlmat*zn/(transpose(transpose(Sn_xlmat)*Sn_xlmat*zn)*(transpose(Sn_xlmat)*Sn_xlmat*zn)))*psii;  
    dak = alph*Sn_xmat*zn; % perturbation of area function, if unconstrained
    dakD = 0; daikD = 0; % initalization of constraint terms
    
    if carea ~= 0 % if kinetic energy constraint on area function is enabled
    
        dpdamat = zeros(size(akp1mat));
        dpmat = diff(akp1mat,1,2);
        dpdamat(:,1) = -2*dpmat(:,1);
        dpdamat(:,2:end-1) = 2*dpmat(:,1:end-1)-2*dpmat(:,2:end);
        dpdamat(:,end) = 2*dpmat(:,end);
        dpdamat2 = dpdamat.*repmat(sum(abs(dpdamat).^2,1),[Nt,1]);
        dakD = dpdamat2(:);
        
    end
    
    if cpot ~= 0 % if potential energy constraint is enabled
        
        daikD_mat = 2*(a0-akp1mat)./a0.*repmat(sum(abs((a0-akp1mat)./a0).^2,1),[Nt,1]);
        daikD = daikD_mat(:);
    
    end
 
    akp1 = abs(ak+diag(ak)*((1-carea)*dak+carea*dakD+cpot*daikD)); % Total perturbation to apply to the previous area function
    Ind = find(akp1<=0.1e-4); % Limit variation in constriction regions (cf Bunton and Story, Estimation of VT area functions in childre based on measurement of lip termination area and inverse acoustic mapping, ICA 2013)
    akp1(Ind) = exp(log(ak(Ind)) + log(dak(Ind)+1)); % New area function
    akp1mat = reshape(akp1, Nt, Ntrame); % Turn area functions into matrix when there are several temporal frames
    if norm_lips; akp1mat(1,:) = a0(1,:); end; % if lip aperture is imposed
    akp1 = akp1mat(:); 
    ak = akp1;   
    
    
    if dl % Compute the new length function
        
        dlkD = 0; dlikD = 0;
        
        if carea ~= 0
        
            dldamat = zeros(size(lkp1mat)); 
            dlmat = diff(cumsum(lkp1mat,1),1,2);
            dldamat(:,1) = -2*dlmat(:,1);
            dldamat(:,2:end-1) = 2*dlmat(:,1:end-1)-2*dlmat(:,2:end);
            dldamat(:,end) = 2*dlmat(:,end);
            dldamat2 = dldamat.*repmat(sum(abs(dldamat).^2,1),[Nt,1]);
            dlkD = dldamat2(:);
        
        end
        
        if cpot ~= 0
            
            dlikD_mat = 2*(l0-lkp1mat).*repmat(sum(abs((l0-lkp1mat)).^2,1),[Nt,1]);
            dlikD = dlikD_mat(:);
            
        end
        
        xi_vec = ((1-cleng)*Sn_xlmat*zn+cleng*dlkD+cpot*dlikD)*alphl;
        Xi_mat = diag(1./(1+xi_vec));
        lkp1 = abs(Xi_mat*lk(:));
%         lkp1(lkp1*40 > 0.2) = 0.2/40; % uncomment to avoid unrealistically long vocal tracts
        lk = lkp1;
        
    end
    
    lkp1mat = reshape(lkp1, Nt, Ntrame);

    % Inialization
    fnest = [];
    Sn_xmat = zeros(length(a0vec),length(ftargvec));
    Sn_xlmat = Sn_xmat;

    for k = 1:Ntrame % Compute sensitivity function for each frame

        fk = af2form( akp1mat(:,k), lkp1mat(:,k), freq, param ); % Estimate the resoannce frequencies at the current iteration
        fn = sort(fk(1:min(length(fk),Nft)),'ascend');
        Sn_x  = Sensitivity_function_fn_v2( lkp1mat(:,k), akp1mat(:,k), fn, param );

        if dl

            Sn_x_l  = Sensitivity_function_length_fn_v2( lkp1mat(:,k), akp1mat(:,k), fn, param);
            Sn_xlmat((k-1)*Nt+1:k*Nt,(k-1)*Nft+1:k*Nft) = Sn_x_l;

        end

        Sn_xmat((k-1)*Nt+1:k*Nt,(k-1)*Nft+1:k*Nft) = Sn_x;
        fnest = [fnest; fn(:)];       
             
    end
    
    count_loop = count_loop+1;
    err_form(count_loop) = norm((fnest-ftargvec)./fnest,1)*100; % Cost function
    err_form(end) % Uncomment for displaying the cost function at the current iteration
    titer(count_loop) = toc; % get the elapsed time at each iteration
    
end

akmat = flipud(akp1mat); % flipud to make the area function from the glottis to the lips
lkmat = flipud(lkp1mat); % flipud to make the length function from the glottis to the lips
fnest = reshape(fnest,[Nft,Ntrame]);

end

function Sn_x  = Sensitivity_function_fn_v2( l, Af, fn, param )
% Compute the sensitivity function Sn along the vocal tract
%
% Input arguments :
%
% l : length function of the vocal tract
% Af : Area function of the vocal tract
% fn : formant frequency
% param : constant terms 
%
% Output arguments : 
%
% Sn_x : sensitivity function. Sn_x is N*M matrix where N is the number
% formant frequencies (length of fn), and M is the length of Af
%
% Benjamin Elie, july 2013

% Size of the sensitivity function
Naf = length(Af);
Nfn = length(fn);
Sn_x = zeros(Naf, Nfn);
rho = param.rho; cs = param.c_s;

[ P, U ] = af2PU_2( Af, l, fn, param ); % Compute the pressure and volume velocity inside each tubelets at the resoannce frequencies

for k = 1:Nfn
    
    Pn = abs(P(:,k)).^2;
	Un = abs(U(:,k)).^2;

    Sn_x(:,k) = l.*(rho./Af.*Un-Af/rho/cs^2.*Pn)./sum(l.*(rho./Af.*Un+Af/rho/cs^2.*Pn));
    
end

end

function Sn_x  = Sensitivity_function_length_fn_v2( l, Af, fn, param )
% Compute the sensitivity function Sn along the vocal tract for the length
% function
%
% Input arguments :
%
% l : length function of the vocal tract
% Af : Area function of the vocal tract
% fn : formant frequency
% param : constant terms 
%
% Output arguments : 
%
% Sn_x : sensitivity function. Sn_x is N*M matrix where N is the number
% formant frequencies (length of fn), and M is the length of Af
%
% Benjamin Elie, july 2013

% Size of the sensitivity function
Naf = length(Af);
Nfn = length(fn);
Sn_x = zeros(Naf, Nfn);
rho = param.rho; cs = param.c_s;

[ P, U ] = af2PU_2( Af, l, fn, param ); % Compute the pressure and volume velocity inside each tubelets at the resoannce frequencies

for k = 1:Nfn
    
    Pn = abs(P(:,k)).^2;
	Un = abs(U(:,k)).^2;

    Sn_x(:,k) = l.*(rho./Af.*Un+Af/rho/cs^2.*Pn)./sum(l.*(rho./Af.*Un+Af/rho/cs^2.*Pn));
    
end

end

function fk = af2form( Avt, lvt, freq, param )
% Compute the transfer function and the impulse response of the vocal tract
% defined by area function Avt and lvt. The resonance frequencies are
% estimated via ESPRIT. They are the pole frequencies of the impulse
% response of the vocal tract. The constant values are stocked in
% param.
% The transfer function is computed using the chain paradigm by Sondhi and
% Schroeter (A hybrid time-frequency domain articulatory speech
% synthesizer, IEEE TASSP, 1987).
% 
% !!THE AREA FUNCTION SHOULD BE FROM THE LIPS TO THE GLOTTIS!!
%
% Input arguments :
%
% Avt : area function of the vocal tract
% lvt : length of the tubelets
% freq : frequency vector
% param : acoustic constant terms
%
% Output argument : 
%
% fk : formant frequencies
%
% Benjamin Elie October 2014

w = 2*pi*freq; % angular frequency
alp = sqrt(1i*w*param.c1);
bet = 1i*w*param.wo2./((1i*w+param.a)*1i.*w+param.b)+alp;
gam = sqrt((param.a+1i*w)./(bet+1i*w));
sig = gam.*(bet+1i*w);
Avt = Avt(:); lvt = lvt(:); freq = freq(:); % should be column vectors
lw = length(w);

Tf = repmat(eye(2),[1 1 lw]); 
A = zeros(1,1,lw); B = A; C = A; D = A;
Ai = Avt; li = lvt;
% Radiation impedance (Flanagan, Speech Analysis, Synthesis and Perception, 2nd edition, 1972)
Zrad = param.rho*(2*pi*freq).^2/2/pi/param.c_s+1i*8*param.rho*(2*pi*freq)/3/pi^(3/2)*Ai(1)^(-1/2);

for k = 1:length(Avt) % Computation of the chain matrix

    Argh = sig*li(k)/param.c_s;

    A(1,1,:) = cosh(Argh);
    B(1,1,:) = -param.rho*param.c_s/Ai(k)*gam.*sinh(Argh);
    C(1,1,:) = -Ai(k)/param.rho/param.c_s./gam.*sinh(Argh);
    Tn = [ A B; C A ];
    Tf = ProdMat3D( Tf, Tn );        

end
    
A = squeeze(Tf(1,1,:)); A = A(:);
B = squeeze(Tf(1,2,:)); B = B(:);
C = squeeze(Tf(2,1,:)); C = C(:);
D = squeeze(Tf(2,2,:)); D = D(:);

Hf = 1./(-C.*Zrad(:)+A);     

[ himp, fs ] = SpecToWave(Hf, freq);
fk = Esprit( himp,fs );

end

function [ P, U ] = af2PU_2( Avt, lvt, freq, param )
% Compute the pressure and air flow inside the vocal and nasal tracts
% defined by area function A and l. The constant values are stocked in
% param. After the chain matrix paradigm by Sondhi and Schroeter (A hybrid time-frequency domain articulatory speech
% synthesizer, IEEE TASSP, 1987).

w = 2*pi*freq; % Angular frequency
alp = sqrt(1i*w*param.c1);
bet = 1i*w*param.wo2./((1i*w+param.a)*1i.*w+param.b)+alp;
gam = sqrt((param.a+1i*w)./(bet+1i*w));
sig = gam.*(bet+1i*w);
Avt = Avt(:); lvt = lvt(:); freq = freq(:);
P = zeros(length(Avt), length(freq));
U = P;
Ai = Avt; li = lvt;

% Radiation impedance (Flanagan, Speech Analysis, Synthesis and Perception, 2nd edition, 1972)
Zrad = param.rho*(2*pi*freq).^2/2/pi/param.c_s+1i*8*param.rho*(2*pi*freq)/3/pi^(3/2)*Ai(1)^(-1/2);
P1 = Zrad;

if iscolumn(P1)
    
    P1 = transpose(P1);
    
end

U1 = ones(size(w));

for k = 1:length(Avt) % Compute the chain matrix

    Argh = sig*li(k)/param.c_s;

    A = cosh(Argh);
    B = -param.rho*param.c_s/Ai(k)*gam.*sinh(Argh);
    C = -Ai(k)/param.rho/param.c_s./gam.*sinh(Argh);
    det_k = 1./(A.^2-B.*C);
    P(k,:) = ((A(:).*P1(:)-B(:).*U1(:))./det_k(:)).';
    U(k,:) = ((-C(:).*P1(:)+A(:).*U1(:))./det_k(:)).';
    P1 = P(k,:);
    U1 = U(k,:);

end

end

function fk = Esprit( x,fs )
% estimate the frequency of the damped sinusoids that model x

K = 0;
M = min(100, floor(length(x)/2));
R = HklBD( x, M ); % Hankel matrix
[U,V,D] = svd(R); % Singular value decomposition
K = Ester( U ); % Estimates the number of components. Comment if you want to skip this step (K = 12 by default)

if K < 12; K = 12; end; % Just to make sure to not underestimate

Up = U(2:end,1:K);
Um = U(1:end-1,1:K);
Phi = pinv(Um)*Up; % Spectral matrix
z = eig(Phi);   % Eigenvalue decomposition, z are the poles that model the signal
freq = angle(z)/2/pi*fs;
fk = freq(freq>50);

end

function K = Ester( W )
% ESTER algorithm to estimate the number of poles included in the signal,
% after Badeau and al. (A new perturbation analysis for signal enumeration in rotational invariance techniques, IEEE TASP, 2006)
%
% Input argument :
%
% W : matrix of the signal subspace
%
% Output argument :
%
% K : dimension of the signal subspace
%
% Benjamin Elie, March 2010

% maximal dimension of the signal subspace:
pmax = size(W,2);
nu = W(end,1)';
mu = nu;
Wup = W(2:end,:);Wdown = W(1:end-1,:);

Psi_lr = Wdown(:,1)'*Wup(:,1);

Ksi = Wup(:,1)-Wdown(:,1)*Psi_lr;
GrandKsi = Ksi;
Phi = mu*conj(Psi_lr);
E = GrandKsi - 1/(1-norm(nu)^2)*(Wdown(:,1)*nu)*Phi'; 
J = zeros(1, ceil(pmax/2));
J(1) = 1/norm(E,2)^2;

for p = 2:ceil(pmax/2)
    
    nu_transp = W(end,1:p);
    nu = nu_transp';
    nu_pmoins1 = nu(1:end-1);
    mu = nu(end);
    Psi_r = Wdown(:,1:p-1)'*Wup(:,p);
    Psi_l = Wup(:,1:p-1)'*Wdown(:,p);
    Psi_lr = Wdown(:,p)'*Wup(:,p);
    
    Ksi = Wup(:,p)-Wdown(:,1:p-1)*Psi_r-Wdown(:,p)*Psi_lr;
    GrandKsi = [ GrandKsi - Wdown(:,p)*Psi_l', Ksi];
    
    Phi_pmoins1 = Phi + mu*Psi_l;
    Phi = [Phi_pmoins1 ; Psi_r'*nu_pmoins1 + mu*conj(Psi_lr)];
    
    if norm(nu) ~= 1
        
        E = GrandKsi - 1/(1-norm(nu)^2)*(Wdown(:,1:p)*nu)*Phi' ;
        J(p) = 1/norm(E,2)^2;
   
    else 
        
        J(p) = 1;
    
    end
    
end

Jp = J(2:2:end); % Real signal : K is pair

thresh_max = max(Jp);
Thresh = thresh_max/10;
        
ind = find(Jp >= Thresh,1,'last'); % find the last value of Jp above the threshold
K = 2*ind; 

end

function [ x, fs ] = SpecToWave( y, f )
% Spectrum to waveform conversion
%
% Input arguments : 
%
% y : complex spectrum
% f : frequency vector
%
% Output arguments
% 
% x : waveform
% fs : sampling frequency
%
% Benjamin Elie October 2010

if isinf(y(1)) || isnan(y(1)) % To avoid singularities
    
    y(1) = y(2);
    
end

y2 = reshape(y,1,length(y));
N = length(y);

df = f(2)-f(1); % frequency step
y2 = [y2 conj(y2(end:-1:2))];

x = ifft(y2,'symmetric');
NIfft = length(x); 
x = NIfft*x; % amplitude normalization
Ls = ceil(length(x)/2)+1;
x = x(1:Ls);
fs = df*2*(N-1); % sampling frequency

end

function C = ProdMat3D( A, B )
% Compute the product of 3D matrices along a single dimension

[ l1, l2, l3 ] = size(A);
C = zeros(size(A));

if l1 == 2 && l2 == 2
    
    C11 = zeros(1,1,l3); C12 = C11; C21 = C11; C22 = C11;
    a11 = squeeze(A(1,1,:)); a12 = squeeze(A(1,2,:)); a21 = squeeze(A(2,1,:)); a22 = squeeze(A(2,2,:));
    b11 = squeeze(B(1,1,:)); b12 = squeeze(B(1,2,:)); b21 = squeeze(B(2,1,:)); b22 = squeeze(B(2,2,:));

    C11(1,1,:) = a11(:).*b11(:) + a12(:).*b21(:); C21(1,1,:) = a21(:).*b11(:)+a22(:).*b21(:);
	C12(1,1,:) = a11(:).*b12(:) + a12(:).*b22(:); C22(1,1,:) = a21(:).*b12(:)+a22(:).*b22(:);
    C = [ C11 C12; C21 C22 ];
    
else

    for k = 1:l3
        
        C(:,:,k) = A(:,:,k)*B(:,:,k);
        
    end
    
end

end

function HH = HklBD( sp, Ntot )
% Compute Hankel matrix for Esprit

Nl = length(sp) - Ntot +1 ; 
Nt = fix(Nl/Ntot); 
HH = zeros(Ntot);


for k = 1:Nt
    
    deb = (k-1)*Ntot+1;
    fin = deb+2*Ntot-1;
    x = sp(deb:fin);
    H = hankel(x(1:Ntot),x(Ntot:end));
    HH = HH + H*H';
    
end

deb = Nt*Ntot+1;
x = sp(deb:end);
H = hankel(x(1:Ntot),x(Ntot:end));
HH = HH + H*H';


end